/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/*
 RFC 1055 SLIP
*/

#include <stdlib.h> 
#include <stdio.h>
#include <string.h>
#include "config.h"
#include "tcpvars.h"
#include "ioconfig.h"

#ifdef USESIMPLESLIP

#include "ip.h"
#include "net.h"
#include "slip.h"
#include "tcpx.h"

/* debug mallocs */
#define SLIPM 0

/* tx one slip character. */

void sliptxone(c)
unsigned int c;
{
	
	if(c == FRAME) {
		PUTDATA(NOTFRAME1);
		PUTDATA(NOTFRAME2);
		return;
	}
	if(c == NOTFRAME1) {
		PUTDATA(NOTFRAME1);
		PUTDATA(NOTFRAME3);
	} else {
		PUTDATA(c);
	}
}

/* rx one slip character */
int sliprxone(VOIDFIX)
{
	unsigned char c;
	unsigned char d;

	c = GETDATA;
	if(c == FRAME) return -1;
	if(c == NOTFRAME1) {
		d = GETDATA;
		if(d == NOTFRAME2) return(FRAME);
		if(d == NOTFRAME3) return(NOTFRAME1);
		return(-1);
	}
	return(c);
}


/* sliptx */
void sliptx(bufnum)
int bufnum;
{
	unsigned int i;
	unsigned char *buf;

	buf = bufptrs[bufnum];  
	txrxflags[bufnum] = ((txrxflags[bufnum] & (~READY)) | BUSY) & 0xff;
#if DEBUGSLIP
	printf("\n\nTX %i\n", bufnum);
#endif
	PUTDATA(FRAME);
	for(i = 0; i < rxtxpointer[bufnum]; i++) {
		sliptxone(buf[i]);
#if DEBUGSLIP
		printf("%2.2x ",buf[i]&0xff);
		if(i == 19 || i == 39 || i == 59 || i == 79) printf("\n");
#endif
#ifdef IOPOLL
		if((i & 0x1f) != 0x00) IOPOLL /* send as we go, but don't eat ALL the cpu */
#endif
	}
	PUTDATA(FRAME);
#if DEBUGSLIP
	printf("\n");
#endif
	/* 
	Free this buffer
	*/
	freebuffer(bufnum);
#ifdef IOPOLL
	IOPOLL
	/* flush the tx buffers */
	while (CHECKOUT > 0) {
		IOPOLL
	}
#endif

}

/* sliprx, snag in scheduled chunks if needbe */
void sliprx(VOIDFIX)
{
	register int i;
	int j;
	int bufnum = -1;
	int freebuf = -1;
	register unsigned char *buf;

	/* what buffer is in-use or free? */
#ifdef IOPOLL
	IOPOLL
#endif
	for(i = 0; i < BUFFERS; i++) {
#ifdef IOPOLL
		IOPOLL
#endif
		if((txrxflags[i] & 0xff) == SWALLOWING) {
			bufnum = i;
		}
		if((txrxflags[i] & 0xff) == 0x00) {
			freebuf = i;
		}
	}
	if((bufnum > -1) || (freebuf > -1)) {
		if(bufnum < 0) {
			/* new buffer */
			bufnum = freebuf;
			bufptrs[bufnum] = malloc(MAXMTUMRU);
			if(bufptrs[bufnum] == NULL) {
				/* ran out of RAM, so we have to wait :-( */
#if SLIPM
				printf("slip.c: %i No RAM\n",__LINE__);
#endif
				return;
			}
			txrxflags[bufnum] = SWALLOWING;
			rxtxpointer[bufnum] = 0;
		}
		/* we have now the buffer and it's pointers */
		/* *PROBLEM*
			if the buffer overflows during a ping flood
			we end up with *ALOT* of garbage and no replies!
			so lets gulp this bastard *fast* in here!
		*/
#ifdef IOPOLL
		IOPOLL
#endif
		j = CHECKIN;
		while(j > 0) {
#ifdef IOPOLL
			IOPOLL
#endif
			i = sliprxone();
			if(i == -1) {
				if(rxtxpointer[bufnum] < 20) {
					/* assume line noise or 
							extra EOF char */
					rxtxpointer[bufnum] = 0x00;
#ifdef IOPOLL
					IOPOLL
#endif
				} else {
					/* end of frame */
					txrxflags[bufnum] = SWALLOWED;
					RECEIVED(bufnum); /* process it */
					/* something to tx? */
					if((txrxflags[bufnum] & (ISTX | READY)) == (ISTX | READY)) {
						/* send it out, this reduces latency */
						sliptx(bufnum);
					}
					j = -1; /* bail out of this loop */
#ifdef IOPOLL
					IOPOLL
#endif
				}
			} else {
#ifdef IOPOLL
				IOPOLL
#endif
				/* some data, save it. */
				buf = bufptrs[bufnum] + rxtxpointer[bufnum];
				buf[0] = i & 0xff;
				rxtxpointer[bufnum]++;
#ifdef IOPOLL
				IOPOLL
#endif
				j = CHECKIN;
				if(rxtxpointer[bufnum] > mrucur) {
					/* buffer has filled!! */
					/* dump the fucker */
					rxtxpointer[bufnum] = 0;
					zapbuffers();
					j = -1;
				}
			}
#ifdef KEYSBAD
#ifdef IOPOLL
			if(fixkbhit()) {
				j = -1; /* bail out, we need to service the keyboard! */
			}
#endif
#endif
		}
	} /* else there was no free buffer */
#ifdef IOPOLL
	IOPOLL
#endif
}

/* sliprxtx */

void sliprxtx(VOIDFIX)
{
	int i;
	int gorx = 0;

	/* walk thru all tx buffers */
	for(i = BUFFERS - 1; i > 0; i--) {
		/* something to tx? */
#ifdef IOPOLL
		IOPOLL
#endif
		if((txrxflags[i] & (ISTX | READY)) == (ISTX | READY)) {
			/* send it out */
			sliptx(i);
		}
#if HANDLEFRAGS
	}
	/* process fragments and anything else swallowed */
	PROCFRAGS
#else
		if((txrxflags[i] & (SWALLOWED)) == (SWALLOWED)) {
			gorx = 1;
			RECEIVED(i);
		}
	}       
#endif /* HANDLEFRAGS */
#ifdef IOPOLL
	IOPOLL
#endif
	if(gorx == 0) sliprx();
}


#else

void notusingslip(VOIDFIX)
{
	return(0);
}

#endif /* USESIMPLESLIP */


/* tell link parser who we are and where we belong

0200 libznet.lib slip.obj xxLIBRARYxx

*/

