/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/* tcp/udp socket code */


#include <stdlib.h> 
#include <stdio.h>
#include <string.h>
#include "tcpvars.h"
#include "in.h"
#include "netinet.h"
#include "socket.h"
#include "ioconfig.h"
#include "tcp.h"
#include "timers.h"
#include "slip.h"
#include "tcpx.h"
#include "udpx.h"
#include "compat.h"

#define DEBUGD 0
#define SOCKM 0

void sendsocket(s)
int s;
{

	register unsigned char *sktp;
	
#if DEBUGD
	printf("sendsocket %i, %i data bytes ",s, sockdlen[s]);
#endif
	sktp = SOKAT(s);
	/* bad, assumes tcp, but we don't care... */
	if(sktp[STTYPE(0, 0)] == SOCK_STREAM) {
		fasttcppack(s, (sktp[BUFSTA(0, 0)] & 0xff), (sktp[SOCKST(0, 0)] & 0xff), sockdlen[s], wdatapointers[s], 0);
	}
}

/* I'm not 100% sure that the time works properly under hi-tech, 
   especially with it's long type oddness... 
   
   However so far it seems to be OK.
 */
void check_sockets(VOIDFIX)
{
	register unsigned char *sktp;
	int i;
	int mss;
	unsigned long now;

	inc32word((unsigned char *)&isnclock);
	for(i = BUFFERS - 1; i > 0; i--) {
		now = CHK_TIME1;
		/* scheduled combined writes */
		
		sktp = SOKAT(i);
		if((sktp[SOCKST(0, 0)] != STATE_CLOSED) && (sktp[SOCKST(0, 0)] != STATE_LISTEN)) {
			if((now > timertx[i]) && (sockdlen[i] == 0) && (wbufcount[i] != 0)) {
				/* spew a waiting build packet! */
				mss = ((sktp[SOKMSS(0, 0)] * 256) + sktp[SOKMSS(0, 1)]) - 24; /* 24 is our standard header length */
				sockdlen[i] = wbufcount[i];
				if(sockdlen[i] > mss) sockdlen[i] = mss;
				errortx[i] = -1;
				sockexpecting[i] = 1;
			}
			if((now>timertx[i]) && ((sockdlen[i] != 0) || (sockexpecting[i] != 0))) {
				errortx[i]++;
				if(errortx[i] > 10) {
					/* socket timed out! ignorantly close it. */
					nuke_tcp_socket(i);
					sockexpecting[i] = 0;
					sockdlen[i] = 0;
				} else {
					if((sktp[SOCKST(0, 0)] != STATE_SYN_SENT) && (sktp[SOCKST(0, 0)] != STATE_SYN_RECEIVED)) {
						sktp[BUFSTA(0, 0)] = PSH | ACK;
					}
					timertx[i] = CHK_TIME1 + (unsigned long)(loopcali * RETRYTIME);
					sendsocket(i);
				}
			}
		}
	}
}

void schedulestuff(VOIDFIX)
{
	check_sockets();
	TRANSPORT
}


int writetcpbytes(fd, bytes, len)
int fd;
unsigned char *bytes;
unsigned int len;
{
	register unsigned char *sktp;
	unsigned char *to;
	int count;
	int mss;
	int status = 0;

	schedulestuff();

	sktp = SOKAT(fd);
	if(sktp[SOCKST(0, 0)] != STATE_ESTABLISHED) {
		status = -1;
	} else {
		if(wbufcount[fd] == 0) {
			timertx[fd] = add32word(CHK_TIME1, (long)(BUFFERS * 2));
		}
		status = len;
		if(status > (MAXTAILW - wbufcount[fd])) {
			status = MAXTAILW - wbufcount[fd];
		}
		count = wbufcount[fd];
		wbufcount[fd] += status;
		to = wdatapointers[fd] + count;
		/* copy bytes */
		memmove(to, bytes, status);
	}
	if((sockdlen[fd] == 0) && (wbufcount[fd] != 0) && (sockexpecting[fd] == 0)) {
		mss = ((sktp[SOKMSS(0, 0)] * 256) + sktp[SOKMSS(0, 1)]) - 24; /* 24 is our standard header length */
		sktp[BUFSTA(0, 0)] = PSH | ACK;
		sockdlen[fd] = wbufcount[fd];
		if(sockdlen[fd] > mss) sockdlen[fd] = mss;
		sockexpecting[fd] = 1;
		timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME); 
		/* fire out this packet _NOW_ */
		sendsocket(fd);
	}
	return(status);
}

/* open, returns handle */
int sockconnect(fd, remoteport, destip)
int fd;
unsigned int remoteport;
unsigned char *destip;
{
	/* 
	fd points to the socket slot here. 
	I'm not going to bother storing the fd twice, 
	so it will return the fd right back upon success. 
	This simplifies alot of things.
	A higher layer (if implemented, MUX for example) 
	should assign it a different one anyhow.
	*/
	register unsigned char *sktp;

	sktp = SOKAT(fd);
	if (sktp[SOCKST(0, 0)] != STATE_NOTINIT) {
		fd = -1;
	} else {
		rdatapointers[fd] = malloc(MAXTAILR);
		if(rdatapointers[fd] == NULL) {
#if SOCKM
			printf("socket.c: %i No RAM\n", __LINE__);
#endif
			return(-1);
		}
		wdatapointers[fd] = malloc(MAXTAILW);
		if(wdatapointers[fd] == NULL) {
			free(rdatapointers[fd]);
#if SOCKM
			printf("socket.c: %i No RAM\n", __LINE__);
#endif
			return(-1);
		}
		switch(sktp[STTYPE(0, 0)]) {
			case SOCK_STREAM: /* tcp */
				fd = tcp_open(fd, remoteport, destip);
				/* now we wait untill such a time that 
				we eiter time out, or get a connection. 
				It really needs something accurate for a timer.
				60hz would be wonderful.
				*/
				if(fd > 0) {
					do {
						schedulestuff();
					} while ((sktp[SOCKST(0, 0)] != STATE_CLOSED) && (sktp[SOCKST(0, 0)] != STATE_ESTABLISHED));
					if(sktp[SOCKST(0, 0)] != STATE_ESTABLISHED) {
						fd = -1;
					}
				}
				break;
				
			default:
				schedulestuff();
				fd = -1;
				break;
		}
	}
	return(fd);
}

/* write */
int sockwrite(fd, sokbuf, len)
int fd;
void *sokbuf;
unsigned int len;
{

	register unsigned char *sktp;
	unsigned char *xbuf;
	int j;
	int status = -1;

	schedulestuff();

	if((fd < 1) || (fd > BUFFERS)) {
		status = -2;
	} else {
		sktp = SOKAT(fd);
		switch(sktp[STTYPE(0, 0)]) {
			case SOCK_STREAM: /* tcp */
				if(sktp[SOCKST(0, 0)] == STATE_ESTABLISHED) {
					j = 0;
					status = 0;
					xbuf = (unsigned char *)sokbuf;
					while(len > 0) {
						j = writetcpbytes(fd, xbuf, len);
						status += j;
						xbuf += j;
						len -= j;
						if(j < 0) {
							len = 0;
						}
					}
				}
				schedulestuff();
				break;

			default:
				break;
		}
	}
	return(status);
}

/* read */
int sockread(fd, sokbuf, len)
int fd;
void *sokbuf;
unsigned int len;
{
	/* sneak in some tcp/ip handling here */
	register unsigned char *sktp;
	unsigned char *bleh;
	unsigned char *here;
	int i;
	int mss;
	int status=-1;

	schedulestuff();

	if((fd < 1) || (fd > BUFFERS)) {
		status = -2;
	} else {
		sktp = SOKAT(fd);
		switch(sktp[STTYPE(0, 0)]) {
			case SOCK_STREAM: /* tcp */
				if(rbufcount[fd] > 0) { /* drain socket */
					/* if there's data here, read it */
					/* for now also our sockets are non-blocking */
					bleh = sokbuf;
					here = rdatapointers[fd];
					for(status = 0; ((status < len) && (rbufcount[fd] != 0)); status++) {
						i = rbuftail[fd];
						bleh[status] = here[i];
						rbufcount[fd]--;
						rbuftail[fd]++;
						if(rbuftail[fd] == MAXTAILR) rbuftail[fd] = 0;
					}
				} else {
					if(sktp[SOCKST(0, 0)] == STATE_ESTABLISHED) {
						status = 0; /* keep socket alive! *till really closed* */
					}
				}
				if(sktp[SOCKST(0, 0)] == STATE_ESTABLISHED) {
					/* extra data push if need be */
					if((sockdlen[fd] == 0) && (wbufcount[fd] > 0)) {
						mss = ((sktp[SOKMSS(0, 0)] * 256) + sktp[SOKMSS(0, 1)]) - 24; /* 24 is our standard header length */
						sktp[BUFSTA(0, 0)] = PSH | ACK;
						sockdlen[fd] = wbufcount[fd];
						if(sockdlen[fd] > mss) sockdlen[fd] = mss;
						errortx[fd] = 0;
						timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
						sockexpecting[fd] = 1;
						sendsocket(fd);
					} else {
						/* advertize our new depth, this should push streaming better */
						if((sockdlen[fd] == 0) && (rbufcount[fd] < (MAXTAILR / 2)) && (status > 0)) {
							sktp[BUFSTA(0, 0)] = ACK;
							errortx[fd] = 0;
							timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
							sendsocket(fd);
						}
					}
				}
				break;
			
			default:
				break;
		}
	}
	return(status);
}

/* for fflush() */
void flushsocket(fd)
int fd;
{
	register unsigned char *sktp;
	
	if((fd < 1) || (fd > BUFFERS)) return;
	
	sktp=SOKAT(fd);
	while(((sktp[SOCKST(0, 0)] == STATE_ESTABLISHED) || (sktp[SOCKST(0, 0)] == STATE_CLOSING)) && ((wbufcount[fd] != 0) || (sockdlen[fd] != 0))) {
		schedulestuff();
	}
}

/* closes socket. WAITS TILL FULLY TIMED OUT! */
int sockAclose(fd)
int fd;
{
	register unsigned char *sktp;
	
	if((fd < 1) || (fd > BUFFERS)) return 0;

	sktp = SOKAT(fd);
	if(sktp[SOCKST(0, 0)] < STATE_ESTABLISHED) {
		nuke_tcp_socket(fd);
	}
	if(sktp[SOCKST(0, 0)] == STATE_ESTABLISHED || sktp[SOCKST(0, 0)] == STATE_CLOSING) {
		/* drain write buffer, if we can */
		flushsocket(fd);
		if(sktp[SOCKST(0, 0)] == STATE_CLOSING) {
			/* send a FIN packet with ack set, state changes here to last ack */
			sktp[SOCKST(0, 0)] = STATE_LAST_ACK;
			sockexpecting[fd] = 0;
			sktp[BUFSTA(0, 0)] = ACK | FIN;
		} else {
			/* send plain fin packet */
			sktp[BUFSTA(0, 0)] = FIN;
			sktp[SOCKST(0, 0)] = STATE_FINWAIT1;
			sockexpecting[fd] = 1;
		}
		errortx[fd] = 0;
		timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
		sendsocket(fd);
	}
domore:
	schedulestuff(); /* make this happen atleast ONCE */
	if((sktp[SOCKST(0, 0)] != STATE_LISTEN) && (sktp[SOCKST(0, 0)] != STATE_CLOSED) && (sktp[SOCKST(0, 0)] != STATE_TIME_WAIT) && (sktp[SOCKST(0, 0)] != STATE_LAST_ACK)) goto domore;
	nuke_tcp_socket(fd);
	return 0;
}

/* closes the socket, but checks for binded sockets and kills those first */
int sockclose(fd)
int fd;
{
	register unsigned char *sktp;
	register unsigned char *foo;
	int i;
	int socks;
	int kill;
	
	if((fd < 1) || (fd > BUFFERS)) return 0;
	sktp = SOKAT(fd);
	/* 
	
	If this is a bind() port, we need to kill allocated sockets, then fall thru. 
	All sockets are killed without prejudice.
	
	*/
	if(sktp[SOCKST(0, 0)] == STATE_BOUND) {
		foo = rdatapointers[fd];
		socks = foo[0] & 0xff;
		i = 0;
		while(i < socks) {
			/* remove parent flag and kill */
			i++;
			kill=foo[i] & 0xff;
			sktp = SOKAT(kill);
			sktp[SPARET(0,0)] = 0x00;
			sockAclose(kill); /* don't care on the return value */
		}
	}
	/* Fall thru and close anything, including our bind socket. */
	kill = sockAclose(fd);
	return(kill);
}

/* find a free socket area */
int findfreeslot(type, protocol)
int type;
int protocol;
{
	register unsigned char* foo;
	int i;
	int fd = -1;
	
	for(i = SOKALLOW; i > 0; i--) {
		if((sockets[SOCKST(i,0)] == STATE_CLOSED) && (rbufcount[i] == 0)) {
			fd = i;
		}
	}
	if(fd>0) {
		foo = SOKAT(fd);
		/* secure this socket NOW! */
		memset((void *)foo, 0, SOKDAT);
		foo[SOCKST(0, 0)] = STATE_NOTINIT; /* allocated, but NOT initialized! */
		foo[STTYPE(0, 0)] = type & 0xff;
		foo[STPROT(0, 0)] = protocol & 0xff;
	}
	return(fd);
}

/* setup socket */
int socket(domain, type, protocol)
int domain;
int type;
int protocol;
{
	int fd = -1;
	if(domain == AF_INET) {
		if(type == SOCK_STREAM) {
			if(protocol == IPPROTO_IP) {
				fd = findfreeslot(type, protocol); /* TCP/IP */
			} /* protocol == 0 */
		} /* type == SOCK_STREAM */
#if UDP_ENABLED

		if(type == SOCK_DGRAM) {
			if(protocol == IPPROTO_UDP) {
				fd = findfreeslot(type, protocol); /* UDP/IP */
			}
		} /* type == SOCK_DGRAM */
#endif /* UDP_ENABLED */
	} /* domain == AF_INET */
	return(fd);
}


/* 
----------------------------------------------------------------------------------------------------------------------------------
Anything below this line is untested and/or incomplete stubs. 
----------------------------------------------------------------------------------------------------------------------------------
*/

/*

When writing nonportable apps, use sockconnect() for just a bit smaller code, 
this is merely a wrapper for compatability, doesn't check sanity.
It may even have endianness problems.

*/

int connect(sockfd, serv_addr, addrlen)
int sockfd;
struct sockaddr_in *serv_addr;
int addrlen;
{
	int retval;

	retval = sockconnect(sockfd, (unsigned short)serv_addr->sin_port, (unsigned char *) (&(serv_addr->sin_addr.s_addr)));
	if(retval > 0) retval = 0;
	return(retval);
}


/*
	bind() 
	creates a blank fd that we can stat via listen()
	The second parameter isn't even used because we do not route.
*/
int bind(fd, my_addr, addrlen)
int fd;
struct sockaddr_in *my_addr;
int addrlen;
{

	register unsigned char *sktp;
	
	if((fd > 0) || (fd < BUFFERS + 1)) {
		sktp = SOKAT(fd);       
		if(sktp[SOCKST(0, 0)] == STATE_NOTINIT) {
			/* save the port # */
			sktp[SRCPRT(0, 1)] = (char) ((unsigned short)my_addr->sin_port & 0xff);
			sktp[SRCPRT(0, 0)] = (char) (((unsigned short)my_addr->sin_port / 256 ) & 0xff);
			/* 
			This is a ficticious socket state, 
			allowing less ram via reuse. 
			This allows some fields to be (ab)used for tracking.
			Don't forget to close the bind sockfd
			if your done with it to free resources!
			*/
#if UDP_ENABLED
			/* set the fake state for UDP, because it requires nothing more. */
			if(sktp[STTYPE(0, 0)] == SOCK_DGRAM) {
				if(sktp[STPROT(0, 0)] == IPPROTO_UDP) {
					sktp[SOCKST(0, 0)] = STATE_UDP;
				}
			} else
#endif /* UDP_ENABLED */
			sktp[SOCKST(0, 0)] = STATE_BIND;

			return(0);
		}
	}
	return(-1);
}

/* 
	listen()
	Backlog is going to be a problem on a small system.
	Each open socket consumes ~3.5k... do the math...
	5 max, * 3.5 = 17k
	So be careful. Running out of RAM isn't fun at all.
	3 is this stack's default, changable in the config up to max buffers - 1.
	This is because bind needs an fd too.
	Returns amount able to allocate. (nonstandard?)
*/
int listen(s, backlog)
int s;
int backlog;
{
	register unsigned char* foo;
	register unsigned char *sktp;
	
	int i;
	int slots;
	int result;

	schedulestuff();
	if((s < 1) || (s > BUFFERS)) {
		return(-3);
	}
	if (backlog < 1) backlog = 1;
	if (backlog > MAXBACKLOG) backlog = MAXBACKLOG;
	rdatapointers[s] = malloc(backlog + 1);
	if(rdatapointers[s] == NULL) {
		return(-2);
	}
	slots = 0;
	foo = SOKAT(s);
	foo[WINDOW(0, 0)] = ((MAXPKTSZ) / 256) & 0xff;
	foo[WINDOW(0, 1)] = (MAXPKTSZ) & 0xff;
	foo[SOKMSS(0, 0)] = ((MAXMTUMRU - 40) / 256) & 0xff;
	foo[SOKMSS(0, 1)] = (MAXMTUMRU - 40) & 0xff;
	for (i = 0; i < backlog ; i++) {
		/* grab backlog count sockets, and make them listen, save each one in the pointer queue */
		/* if we can't grab all sockets, truncate. No slots is an error. */
		result = findfreeslot(foo[STTYPE(0, 0)], foo[STPROT(0, 0)]);
		if(result > 0) {
			sktp = SOKAT(result);
			/* Allocate resources, if we can. */
			rdatapointers[result] = malloc(MAXTAILR);
			if(rdatapointers[result] != NULL) {
				wdatapointers[result] = malloc(MAXTAILW);
				if(wdatapointers[result] != NULL) {
					/* copy the socket, fix the state */
					slots++;
					sktp[SPARET(0, 0)] = s & 0xff;
					sktp[SPARPO(0, 0)] = slots & 0xff;
					nuke_tcp_socket(result);
				} else {
					/* Reclaim resources. */
					free(rdatapointers[result]);
					sktp[SOCKST(0, 0)] = STATE_CLOSED;
				}
			} else {
				/* Reclaim resources. */
				sktp[SOCKST(0, 0)] = STATE_CLOSED;
			}
		}
	}
	sktp = rdatapointers[s];
	sktp[0] = slots & 0xff;
	if(slots == 0) {
		free(rdatapointers[s]);
		return(-1);
	} else {
		foo[SOCKST(0, 0)] = STATE_BOUND;
	}
	return(slots);
}

/*
	sockstatus()
	check if socket() is at STATE_ESTABLISHED
*/
int sockstatus(s)
int s;
{
	register unsigned char *sktp;

	schedulestuff();
	if((s > 0) || (s < BUFFERS + 1)) {
		sktp = SOKAT(s);        
		return(sktp[SOCKST(0, 0)] & 0xff);
	}
	return(-1);
}

/*
	bindcount()
	Returns the count of nonestablished connections on a bound socket.
*/

int bindcount(s)
int s;
{
	register unsigned char *sktp;
	register unsigned char *foo;
	int i;
	int j;
	int count;
	int retval;
	
	schedulestuff();
	if((s < 1) || (s > BUFFERS)) {
		return(-1);
	}
	sktp = SOKAT(s);
	if(sktp[SOCKST(0, 0)] != STATE_BOUND) {
		 return(-2);
	}
	sktp = rdatapointers[s];
	count = sktp[0] & 0xff;
	i = 0;
	retval = 0;
	while(i < count) {
		i++;
		j = sktp[i] & 0xff;
		foo = SOKAT(j);
		if((j != 0x00) && (foo[SOCKST(0,0)] != STATE_ESTABLISHED)) retval++; 
	}
	return(retval);
}

/*
	acceptready()
	Returns if ANY socket in the backlog can be accepted.
*/

int acceptready(s)
int s;
{
	register unsigned char *sktp;
	register unsigned char *foo;
	int i;
	int j;
	int count;
	
	schedulestuff();
	if((s < 1) || (s > BUFFERS)) {
		return(-1);
	}

	sktp = SOKAT(s);
	if(sktp[SOCKST(0, 0)] != STATE_BOUND) {
		return(-2);
	}

	sktp = rdatapointers[s];
	count = sktp[0] &0xff;
	i = 0;
	while(i < count) {
		i++;
		j = sktp[i] & 0xff;
		foo = SOKAT(j);
		if((j != 0x00) && (foo[SOCKST(0,0)] == STATE_ESTABLISHED)) return(1);
	}
	return(0); /* none ready */  
}

/*
	accept()
	Unfortunately, this one has to block.
	Returns the fd to the socket that connected or error.
*/

int accept(s, addr, addrlen)
int s;
struct sockaddr_in *addr;
int *addrlen;
{
	register unsigned char *sktp;
	int this;
	int sock;
	int status;
	int count;
	
	schedulestuff();
	if((s < 1) || (s > BUFFERS)) {
		return(-1);
	}

	sktp = SOKAT(s);
	if(sktp[SOCKST(0, 0)] != STATE_BOUND) {
		return(-2);
	}

	sock = -1;
	if(bindcount(s) < 0) {
		return(-1); /* error */
	}
	
	/* dig out the fds from the listening sockets */
	sktp = rdatapointers[s];
	count = sktp[0] & 0xff;
	while(sock == -1 ) {
		this = 0;
		while(this < count) {
			this++;
			status = sockstatus(sktp[this]);
			if((status == STATE_ESTABLISHED) && (sktp[this] > 0)) {
				sock = sktp[this];
				sktp[this] = 0x00; /* now dead, out of the ring */
				/* fill in the address/port info for return */
				sktp = SOKAT(sock);
				addr->sin_family = sktp[STTYPE(0, 0)];
				addr->sin_port = sktp[DESPRT(0, 0)] + (sktp[DESPRT(0, 1)] * 256);
				snag32((unsigned char *)&sktp[IPDEST(0,0)], (unsigned char *)&addr->sin_addr);
				this = count + 1;
			}
		}
	}               
	/* return the listening socket # */
	return(sock);
}

#if 0
#if UDP_ENABLED
/*

UDP send is unique in the fact that it is deleivered instantly.
A check on the TX queue is needed to verify there is room so that
it does not block. UDP is not buffered at all in this implementation.
On some platforms that have no scheduling, there is no TX queue.
This is because slip verifies the buffer as empty.

*/
int sendto(s, msg, len, flags, to, tolen)
int s;
const void *msg;
size_t len;
int flags;
const struct sockaddr *to;
socklen_t tolen;
{
	register unsigned char *sktp;
	unsigned char *buffer;
	unsigned char *ptr;
	unsigned char *pts;
	int i;
	int freebuf;
	
	schedulestuff();



	if((s > 1) || (s <= BUFFERS)) {
		sktp = SOKAT(s);
		/* assumes UDP */
		if(sktp[STTYPE(0, 0)] == SOCK_DGRAM) {
			/* we do not fragment, so if it can't fit in the message size, error! */
			if(len <= (size_t)(CHECKUP)) {
				/* we can xmit w/out blocking... build a UDP packet and ship it */
				for(i = 1; i < BUFFERS; i++) {
					if((txrxflags[i] & 0xff) == 0x00) freebuf = i;
				}
				if(freebuf < 0) {
					goto wouldblock; /* out of build buffers */
				}
				txrxflags[freebuf] = (ISTX | FILLBUSY); /* lock the buffer */
				buffer = bufptrs[freebuf];
				ptr = to->sin_addr;
				/* lay out ip */
				hi = createip(buffer, ptr, 0x06);
				ptr = to->sin_port;
				ht = createudp(buffer, ptr, datlen, datptr, sktp);
				/* calculate sums */
				buffer = bufptrs[freebuf];
				udpsumming(buffer); /* needs to check the option for summing as well */

				/* set the size of the entire package */
				rxtxpointer[freebuf] = hi + ht;
				/* toss it in the tx que */
				txrxflags[freebuf] = (ISTX | READY | SENDNOCARE) & 0xff;
			} else {
				goto wouldblock; /* no nic buffer free */
			}
		}
	}
	errno = ENOSYS;
	return(-1);
wouldblock:
	errno = EWOULDBLOCK;
	return(-1);
}

int recvfrom(s,  buf, len, flags, from, fromlen)
int  s;
void  *buf;
size_t len;
int flags;
struct sockaddr *from;
socklen_t *fromlen;
{
	register unsigned char *sktp;
	unsigned char *here;
	int i;
	int count = 0;
	int pcount;
	schedulestuff();

	if((s > 0) || (s <= BUFFERS)) {

		sktp = SOKAT(s);
		/* assumes UDP */
		if(sktp[STTYPE(0, 0)] == SOCK_DGRAM) {
			/* data? */
			if(rbufcount[s] == 0) {
				errno = EWOULDBLOCK;
				return(-1);
			}
			/* slop the content where it needs to go */
			here = rdatapointers[s];
			i = rbuftail[fd]; /* 0 */
			from->sin_family = here[i]; /* family */

			i++; /* 1 (0) */
			if(i == MAXTAILR) i = 0;
			from->sin_addr[0] = here[i]; /* ad 0 */
			i++; /* 2 (1) */
			if(i == MAXTAILR) i = 0;
			from->sin_addr[1] = here[i]; /* ad 1 */
			i++; /* 3 (2) */
			if(i == MAXTAILR) i = 0;
			from->sin_addr[2] = here[i]; /* ad 2 */
			i++; /* 4 (3) */
			if(i == MAXTAILR) i = 0;
			from->sin_addr[3] = here[i]; /* ad 3 */

			i++; /* 5 (0) */
			if(i == MAXTAILR) i = 0;
			from->sin_port[0] = here[i]; /* p 0 */
			i++; /* 6 (1) */
			if(i == MAXTAILR) i = 0;
			from->sin_port[1] = here[i]; /* p 1 */

			i++; /* 7 (0) */
			if(i == MAXTAILR) i = 0; /* skip 0 */
			i++; /* 8 (1) */
			if(i == MAXTAILR) i = 0; /* skip 1 */

			i++; /* 9 (0) */
			if(i == MAXTAILR) i = 0;
			pcount = here[i];	/* l 0 */
			i++; /* 10 (1) */
			if(i == MAXTAILR) i = 0;
			pcount = here[i];	/* l 1 */
			
			i++; /* 11 (0) */
			if(i == MAXTAILR) i = 0; /* c 0 */  
			i++; /* 11 (1) */
			if(i == MAXTAILR) i = 0;  /* c 1 */
			rbuftail[fd] = i;
			rbufcount[fd] -= 11;
			
			count = pcount - 8;
			pcount = count;
			/* increment tail as we go */
			while(count) {
				*msg++ = here[i];
				i++;
				if(i == MAXTAILR) i = 0;
				rbufcount[fd]--;
			}
			return(pcount); /* return count in packet */
		}
	}
	errno = ENOSYS;
	return(-1);
}


#else /* UDP_ENABLED */

int sendto(s, msg, len, flags, to, tolen)
int s;
const void *msg;
size_t len;
int flags;
const struct sockaddr *to;
socklen_t tolen;
{
	errno = ENOSYS;
	return(-1);

}

int recvfrom(s,  buf, len, flags, from, fromlen)
int  s;
void  *buf;
size_t len;
int flags;
struct sockaddr *from;
socklen_t *fromlen;
{
	errno = ENOSYS;
	return(-1);

}

#endif /* UDP_ENABLED */
#endif
/* tell link parser who we are and where we belong

0100 libznet.lib socket.obj xxLIBRARYxx

*/

