/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/*

TCP RFC 0793
Pre-optimized for the compiler via
keeping 0x00 stores together and other tricks to help the C optimizer
I know this makes it harder to follow, but it's worth it! HONEST!

Unfortunately writes are commited instantly due to small TPA constraints.
This may change in the future, providing I get to write a C compiler
that produces less inline asm. Calls may be a bit time expensive, 
but ram is the major issue on smaller systems.

There is a specail case where the stack gets out-of-sync with the incoming stream
due to low speeds. I'm not quite certain how to fix that yet.

*/

#include <stdlib.h> 
#include <stdio.h>
#include <string.h>
#include "tcpvars.h"
#include "in.h"
#include "netinet.h"
#include "socket.h"
#include "ioconfig.h"
#include "chcksum.h"
#include "timers.h"
#include "tcpx.h"
#include "compat.h"
#include "icmp.h"

#if IP_ENABLED
#if TCP_ENABLED
#define DEBUGI 0
#define DEBUGD 0
#define DEBUGM 0
#define DEBUGQUE 0

void nuke_tcp_socket(fd)
int fd;
{
	register unsigned char *sktp;
	register unsigned char *foo;
	int par;
	int i;
	
	sktp = SOKAT(fd);
	if(sktp[SPARET(0,0)] == 0x00) {
		sktp[SOCKST(0, 0)] = STATE_CLOSED;
		free(rdatapointers[fd]);
		free(wdatapointers[fd]);
	} else {
		/* Reclaim this buffer */
		par = sktp[SPARET(0, 0)];
		i = sktp[SPARPO(0, 0)] & 0xff;
		foo = SOKAT(i);
		memcpy(sktp, foo, SOKDAT);
		sktp[SPARET(0, 0)] = par & 0xff;
		sktp[SPARPO(0, 0)] = i & 0xff;
		sktp[SOCKST(0, 0)] = STATE_LISTEN;
		sktp=rdatapointers[par];
		sktp[i] = fd & 0xff; /* reclaimed */
	}
	rbufcount[fd] = 0;
	wbufcount[fd] = 0;
}

void processtcpdata(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char sockstate;
	unsigned char command;
	unsigned char *buffer;
	unsigned long oack;
	unsigned long iseq;
	long diff;
	unsigned char *datapointer;
	unsigned char *ours;
	
#if DEBUGD
	printf("\nprocesstcpdata\n");
#endif
	sktp = SOKAT(sock);
	buffer = bufptrs[bufnum];
	sockstate = sktp[SOCKST(0, 0)];
	command = buffer[iplen + 13];
	oack = ret32word(&sktp[ACKNUM(0, 0)]);
	iseq = ret32word(NUMSEQ(buffer, iplen));
	/* needs to handle 0xffffffff roll over !!!!!! */
	diff = sub32word(oack, iseq);
#if DEBUGD
	printf("DL %4.4x DIFF %8.8lx oack %8.8lx iseq %8.8lx\n", datalen, diff, oack, iseq);
#endif
	if((compare32word(diff, CMPLEQ, (long)(MAXTAILR - rbufcount[sock])))) {
		/* we have enuff buffer space to swallow this data */

		if ( command & SYN ) diff--;
		if ( (compare32word(diff, CMPEQU, 0L)) ) { /* match only *EXACT* packets we expect as we do NOT window! */
			datapointer = buffer + iplen + tcplen - datalen;
			sav32word(&sktp[ACKNUM(0, 0)], add32word(oack, (long)datalen));
			bufferdatamove(datapointer, sock, datalen);
		
			if(sockstate == STATE_FINWAIT1) {

				if ( (command & (FIN | ACK)) == FIN ) {
					/* we only got a fin packet, so we need to change state so we can resync */
					sktp[SOCKST(0, 0)] = STATE_CLOSING;
					sktp[BUFSTA(0,0)] = ACK | FIN;
#if DEBUGD
					printf("STATE_CLOSING(fixup)\n");
#endif
				} else {
					freebuffer(bufnum);
					sockexpecting[sock] = 0;
					sktp[BUFSTA(0, 0)] = ACK;
					sktp[SOCKST(0, 0)] = STATE_FINWAIT2;
#if DEBUGD
					printf("STATE_FINWAIT2\n");
#endif
					return;
				}
			}
			if ( command & FIN ) {
			
				ours = &sktp[ACKNUM(0, 0)];
				inc32word(ours);
#if DEBUGD
				printf("consumed fin.\n");
#endif
				if(sockstate == STATE_FINWAIT2) {
#if DEBUGD
					printf("STATE_TIME_WAIT\n");
#endif
					sktp[SOCKST(0, 0)] = STATE_TIME_WAIT;
					sockexpecting[sock] = 0;
					sktp[BUFSTA(0, 0)] = ACK;
				}
				if(sockstate == STATE_ESTABLISHED) {
					sktp[SOCKST(0, 0)] = STATE_CLOSING;
				}
			}
			/* free the inbound buffers */
			freebuffer(bufnum);
			timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
			sendsocket(sock); /* ACK back */
		} /* 
		  If window didn't match... 
		  Send our current ack reguardless.
		  Hope we get a retransmit of what we want?
		  */

		/* free the inbound buffers */
		else {
			freebuffer(bufnum);
		}
	} else {
		/* else we save the packet for another time, should we send winsize=0? no read buffer left :-( 
		however this doesn't mean we can't send our data out... we'll see what happens with it this way first :O) */
#if DEBUGQUE
		printf("out of buffer? Buffer free %8.8lx, inbound %8.8lx\n", (long)(MAXTAILR - rbufcount[sock]), diff);
#endif
		freebuffer(bufnum);
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		sendsocket(sock);
	}
}


void listenproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char command;
	unsigned char *buffer;
	unsigned char *ours;

#if DEBUGD
	printf("\nlistenproc\n");
#endif
	buffer = bufptrs[bufnum];
	command = buffer[iplen + 13];
	if ( command & SYN ) {
		sktp=SOKAT(sock);
		sktp[SOCKST(0, 0)] = STATE_SYN_RECEIVED;
		snag16(PRTSRC(buffer, iplen), DESPRT(sock, sockets));
		ours = &sktp[ACKNUM(0, 0)];
		snag32(NUMSEQ(buffer, iplen), ours);
		inc32word(ours);
		ours = &sktp[SEQNUM(0, 0)];
		sav32word(ours, isnclock);
		snag32(DESTIP(buffer), &sktp[IPDEST(0, 0)]);
		tcp_synandack(sock);
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		sockexpecting[sock] = 1;
	} else {
		sock = 0;
		sktp = SOKAT(sock);
		snag32(&buffer[12],        &sktp[IPDEST(0, 0)]);
		snag16(&buffer[iplen],     &sktp[DESPRT(0, 0)]);
		snag16(&buffer[iplen + 2], &sktp[SRCPRT(0, 0)]);
		snag32(&buffer[iplen + 4], &sktp[ACKNUM(0, 0)]);
		snag32(&buffer[iplen + 8], &sktp[SEQNUM(0, 0)]);
		tcp_rst(sock); /* use special buffer */
	}
	/* free the inbound buffers */
	freebuffer(bufnum);
}

void syn_sentproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char command;
	unsigned char *buffer;
	unsigned long iack;
	unsigned long iseq;
	unsigned long oseq;

#if DEBUGD
	printf("\nsyn_sentproc\n");
#endif
	buffer = bufptrs[bufnum];
	command = buffer[iplen + 13];
	sktp = SOKAT(sock);
	iack = ret32word(NUMACK(buffer,iplen));
	iseq = ret32word(NUMSEQ(buffer,iplen));
	oseq = ret32word(&sktp[SEQNUM(0, 0)]);
	if ( command & SYN ) {
		inc32word(&sktp[ACKNUM(0, 0)]);
		sktp[BUFSTA(0, 0)] = ACK;
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		if ( (command & ACK) && (iack == (oseq + 1))) {
			sktp[SOCKST(0, 0)] = STATE_ESTABLISHED;
			inc32word(&sktp[SEQNUM(0, 0)]);
			sav32word(&sktp[ACKNUM(0, 0)], (iseq + 1));
			sockexpecting[sock] = 0;
			errortx[sock] = 0;
		} else {
			sktp[SOCKST(0, 0)] = STATE_SYN_RECEIVED;
			errortx[sock] = 0;
		}
	}
	/* free the inbound buffer as we are done with it now */
	freebuffer(bufnum);
	/* task the other side */
	sendsocket(sock);
}

void syn_receivedproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char command;
	unsigned char *buffer;
	unsigned long iack;
	unsigned long oseq;

#if DEBUGD
	printf("\nsyn_receivedproc\n");
#endif
	buffer = bufptrs[bufnum];
	command = buffer[iplen + 13];
	sktp = SOKAT(sock);
	iack = ret32word(NUMACK(buffer, iplen));
	oseq = ret32word(&sktp[SEQNUM(0, 0)]);
	if ( command & SYN ) {
		sktp[BUFSTA(0,0)] = SYN | ACK;
		sendsocket(sock);
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		errortx[sock] = 0;
	}
	if ( (command & ACK) && (iack == (oseq + 1))) {
		sktp[BUFSTA(0,0)] = ACK;
		sendsocket(sock);
		inc32word(&sktp[SEQNUM(0, 0)]);
		sockexpecting[sock] = 0;
		sktp[SOCKST(0,0)] = STATE_ESTABLISHED;
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		errortx[sock] = 0;
	}
	/* free the inbound buffers */
	freebuffer(bufnum);
}

void establishedproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char sockstate;
	unsigned char command;
	unsigned char *buffer;
	unsigned char *front;
	unsigned char *back;
	unsigned long iack;
	unsigned long oack;
	unsigned long iseq;
	unsigned long oseq;
	unsigned long expected;
	int count;
	long diff;


#if DEBUGD
	printf("\nestablishedproc\n");
#endif
	buffer = bufptrs[bufnum];
	sktp = SOKAT(sock);
	sockstate = sktp[SOCKST(0, 0)];
	command = buffer[iplen + 13];
	iack = ret32word(NUMACK(buffer, iplen));
	oack = ret32word(&sktp[ACKNUM(0, 0)]);
	iseq = ret32word(NUMSEQ(buffer, iplen));
	oseq = ret32word(&sktp[SEQNUM(0, 0)]);
	if ( (command & ACK) == 0 ) {
		printf("tcp.c: command bug %2.2x\n",command);
		freebuffer(bufnum);
		return;
	}
	sockexpecting[sock] = 0;
	sktp[BUFSTA(0, 0)] = ACK;
	errortx[sock] = 0;
	/* process ack value in packet */
	count = sockdlen[sock];
	expected = count;
	/* this ugly conglomeration of shit handles 0xffffffff rollover/rollbacks */
	diff = sub32word((add32word(oseq, expected)), iack);
	if ((compare32word(diff, CMPLEQ, expected))) {
		count = expected - diff;
		front = wdatapointers[sock];
		back = wdatapointers[sock] + count;
		wbufcount[sock] -= count;
		sockdlen[sock] -= count;
		sav32word(&sktp[SEQNUM(0, 0)], add32word(oseq, sub32word(diff, expected)));
		sockexpecting[sock] = 0;

#if DEBUGM
printf("%4.4x OCTETS SENT, BUFFERS: %4.4x %4.4x\n",count,wbufcount[sock],sockdlen[sock]);
printf("MEMMOVE -> TO %4.4x FROM %4.4x FOR %4.4x OCTECTS TCP.C, line %i\n", back, front, wbufcount[sock], __LINE__);
#endif
		memmove(front, back, (wbufcount[sock]));
	}
	/*
	we should hit this only if:
	1: we have data
	2: we recieved data
	3: any fin flags set
	*/
	if(((command & ~ACK) != 0) || (wbufcount[sock] != 0) || (datalen != 0)) {
		processtcpdata(bufnum, sock, iplen, tcplen, datalen);
		/* the inbound buffers are already free */
	} else {
		freebuffer(bufnum);
	}
}

void finwait1proc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char sockstate;
	unsigned char command;
	unsigned char *buffer;

#if DEBUGD
	printf("\nfinwait1proc\n");
#endif
	buffer = bufptrs[bufnum];
	sktp = SOKAT(sock);
	sockstate = sktp[SOCKST(0, 0)];
	command = buffer[iplen + 13];
#if DEBUGD
		printf("finwait1proc: command %2.2x\n", command);
#endif
	if((command & FIN) == 0) {
		/* toss it, this is an error */
		freebuffer(bufnum);
		return;
	}

	processtcpdata(bufnum, sock, iplen, tcplen, datalen);
}

void finwait2proc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char sockstate;
	unsigned char command;
	unsigned char *buffer;
	unsigned long iack;
	unsigned long oseq;

#if DEBUGD
	printf("\nfinwait2proc\n");
#endif  
	buffer=bufptrs[bufnum];
	sktp = SOKAT(sock);
	sockstate = sktp[SOCKST(0, 0)];
	command = buffer[iplen + 13];
	iack = ret32word(NUMACK(buffer,iplen));
	oseq = ret32word(&sktp[SEQNUM(0, 0)]);
	errortx[sock] = 0;
	sktp[BUFSTA(0, 0)] = FIN | ACK;
	if((command & FIN) && (compare32word(sub32word(iack, oseq), CMPEQU, 0L))) {
		sktp[SOCKST(0, 0)] = STATE_TIME_WAIT;
		sktp[BUFSTA(0, 0)] = ACK;
	}
	processtcpdata(bufnum, sock, iplen, tcplen, datalen); /* answer it */
	/* the inbound buffers are already free */
}

void closingproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char *buffer;
	unsigned long iack;
	unsigned long oseq;


#if DEBUGD
	printf("\nclosingproc\n");
#endif  
	buffer = bufptrs[bufnum];
	sktp = SOKAT(sock);
	iack = ret32word(NUMACK(buffer,iplen));
	oseq = ret32word(&sktp[SEQNUM(0, 0)]);
	if ( iack == (oseq + 1) ) {
		sktp[SOCKST(0, 0)] = STATE_TIME_WAIT;
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		errortx[sock] = 0;
		sktp[BUFSTA(0,0)] = ACK | FIN;
		freebuffer(bufnum);
		sendsocket(sock);
	} else {
		/* free the inbound buffers */
		freebuffer(bufnum);
	}
}

void last_ackproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	unsigned char *buffer;
	unsigned long iack;
	unsigned long oseq;


#if DEBUGD
	printf("\nlast_ackproc\n");
#endif
	buffer = bufptrs[bufnum];
	sktp = SOKAT(sock);
	iack = ret32word(NUMACK(buffer,iplen));
	oseq = ret32word(&sktp[SEQNUM(0, 0)]);
	if ( iack == (oseq + 1) ) {
		sockexpecting[sock] = 0;
		sockdlen[sock] = 0;
		errortx[sock] = 0;
		nuke_tcp_socket(sock);
	} else {
		sktp[BUFSTA(0, 0)] = ACK | FIN;
		sendsocket(sock);
		timertx[sock] = CHK_TIME1 + (loopcali * RETRYTIME);
		errortx[sock] = 0;
	}
	/* free the inbound buffers */
	freebuffer(bufnum);
}

void time_waitproc(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
#if DEBUGD
	printf("\ntime_waitproc\n");
#endif
	sktp = SOKAT(sock);
	sktp[BUFSTA(0, 0)] = ACK;
	errortx[sock] = 0;
	sendsocket(sock);
	/* free the inbound buffers */
	freebuffer(bufnum);
}

void proctcp(bufnum, sock, iplen, tcplen, datalen)
unsigned int bufnum;
int sock;
int iplen;
int tcplen;
int datalen;
{
	register unsigned char *sktp;
	register unsigned char sockstate;

#if DEBUGD
	printf("proctcp-> ");
#endif
	sktp = SOKAT(sock);
	sockstate = sktp[SOCKST(0, 0)];

	switch(sockstate) {
		case STATE_LISTEN:
			listenproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_SYN_SENT:
			syn_sentproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_SYN_RECEIVED:
			syn_receivedproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_ESTABLISHED:
			establishedproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_FINWAIT1:
			finwait1proc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_FINWAIT2:
			finwait2proc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_CLOSING:
			closingproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_LAST_ACK:
			last_ackproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		case STATE_TIME_WAIT:
			time_waitproc(bufnum, sock, iplen, tcplen, datalen);
			break;

		default:
			freebuffer(bufnum); 
			break;
	}
	return;
}

/* rx TCP RFC 0793 */
void tcp(bufnum, iplen, totallen, tcplen)
unsigned int bufnum;
unsigned int iplen;
unsigned int totallen;
unsigned int tcplen;
{
	register unsigned char *sktp;
	register unsigned char *buffer;
	unsigned int datalen;
	unsigned int sum;
	unsigned int tsum;
	unsigned char a;
	unsigned char b;
	unsigned char fakebuf[13];
	int sock= - 1;

	buffer = bufptrs[bufnum];
	datalen = tcplen - ((buffer[iplen + 12] & 0xf0) / 4);

	fakebuf[ 0] = buffer[12];
	fakebuf[ 1] = buffer[13];
	fakebuf[ 2] = buffer[14];
	fakebuf[ 3] = buffer[15];

	fakebuf[ 4]  = buffer[16];
	fakebuf[ 5]  = buffer[17];
	fakebuf[ 6]  = buffer[18];
	fakebuf[ 7]  = buffer[19];

	fakebuf[ 8] = 0x00;
	fakebuf[ 9] = 0x06;
	fakebuf[10] = (tcplen / 256 ) & 0xff;
	fakebuf[11] = tcplen & 0xff;

#if DEBUGD

	printf("\nTCP ");

#endif
	a = buffer[iplen + 16];
	b = buffer[iplen + 17];
	buffer[iplen + 16] = 0x00;
	buffer[iplen + 17] = 0x00;
	tsum = checksum(fakebuf, 12, &buffer[iplen], tcplen);
	sum = (a * 256) + b;
	/* packet flag type */
	a = buffer[iplen + 13];
	if(sum == tsum) {
		/* sum is good, check the ports */
		sock = matchsocket(&buffer[iplen]);
		if(sock<0) {

#if DEBUGI
			printf("No match on socket ");
#endif
			if((a & (RST)) == 0x00) {

				if((a & (SYN)) == SYN) {
					/* type 3, code 3 port unreachable */
					err_reply(bufnum, 3, 3, 0);
				} else {
					/* send a RST reply or, if this is a FIN, finack it */
					sock = 0;
					sktp = SOKAT(sock);
					snag32(&buffer[12],        &sktp[IPDEST(0, 0)]);
					snag16(&buffer[iplen],     &sktp[DESPRT(0, 0)]);
					snag16(&buffer[iplen + 2], &sktp[SRCPRT(0, 0)]);
					snag32(&buffer[iplen + 4], &sktp[ACKNUM(0, 0)]);
					snag32(&buffer[iplen + 8], &sktp[SEQNUM(0, 0)]);
					if((a & FIN) == FIN) {
						if((a & (FIN | ACK)) == (FIN | ACK)) {
							tcp_fin(sock);
						} else {
							tcp_finack(sock);
						}
					} else {
						tcp_rst(sock); /* use special buffer */
					}
					freebuffer(bufnum);
				} 
			}
		} else {
			sktp = SOKAT(sock);
#if DEBUGI
			printf("socket %i ",sock);
			fflush(stdout);
#endif
			/* socket state */
			b = sktp[SOCKST(0, 0)];
			if((a & 0xc0) != 0x00) {
#if DEBUGI
				printf("Malformed flags ");
#endif
				freebuffer(bufnum);

			} else {
				if((a & RST) == RST) {
					nuke_tcp_socket(sock);
				} else {
					/* passes most header checks, process packet */
					proctcp(bufnum, sock, iplen, tcplen, datalen);
				}
			}
		}
	} else {
#if DEBUGI
		printf("Bad checksum\n");
#endif
		freebuffer(bufnum); /* bad checksum */
	}/* endif sum == tsum */

#if DEBUGI || DEBUGD
			printf("\n");
#endif
	return;
}



#else /* TCP_ENABLED */

/* no tcp available, so.... */
void tcp(bufnum, loh, lop, lod)
unsigned int bufnum
unsigned int loh;
unsigned int lop;
unsigned int lod;
{
	/* type 3, code 2 protocol unreachable */
	err_reply(bufnum, 3, 2, 0);
}
#endif /* TCP_ENABLED */

#else /* IP_ENABLED */
#endif /* IP_ENABLED */

/* tell link parser who we are and where we belong

0700 libznet.lib tcp.obj xxLIBRARYxx

*/

