/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/*

UDP RFC 0768

Portions sponsored by Building 31 Clustering AB, Sweden.

*/

#include <stdlib.h> 
#include <stdio.h>
#include <string.h>
#include "tcpvars.h"
#include "in.h"
#include "netinet.h"
#include "icmp.h"
#include "tcpx.h"
#include "chcksum.h"

#if IP_ENABLED
#if UDP_ENABLED
#define DEBUGUDP 1

/* rx UDP RFC 0768 */

void udp(bufnum, iplen, totallen, udplen)
unsigned int bufnum;
unsigned int iplen;
unsigned int totallen;
unsigned int udplen;
{
	unsigned char *buffer;
	unsigned char *bufflags;
	unsigned char *srcip;
	unsigned char a;
	unsigned char b;
	int sum;
	int usum;
	int sock = -1;
		
	buffer = bufptrs[bufnum];
	if(udplen < 8) goto black_hole;
	srcip = remip + 4 * bufnum;
	bufflags = txrxflags + bufnum;
	a = buffer[iplen + 6];
	b = buffer[iplen + 7];
	sum = (a * 256) + b;
#if DEBUGUDP
	printf("\nUDP ");
#endif

	/* 
	   3 steps:
	   1: Test checksum.
	   2: Match inbound port.
	   3a: If there is no room, black hole the packet.
	   3b: Place package into the queue.
	 */
	/* 1: Test checksum, if there is one.... */
	if(sum != 0) {
		/* there is a checksum */
		buffer[iplen + 6] = 0x00;
		buffer[iplen + 7] = 0x00;
		usum = checksum(NULL, 0, buffer, totallen);
		if(usum == 0) usum=0xffff;
#if DEBUGUDP
		printf("CHECKSUM I:O %4.4x %4.4x ", sum, usum);
#endif
		if(sum != usum) {
#if DEBUGUDP
			printf("BAD ");
#endif
			goto black_hole;

#if DEBUGUDP
		} else {
			printf("OK ");
#endif
		}
		buffer[iplen + 6] = a;
		buffer[iplen + 7] = b;
#if DEBUGUDP
	} else {
		printf("NO CHECKSUM, OK ");
#endif
	}
	/* 2: Match inbound port. */
	sock = matchsocket(&buffer[iplen]);
	if(sock < 0) {
		/* type 3, code 3 port unreachable */
		err_reply(bufnum, 3, 3, 0);
#if DEBUGUDP
		printf("NO SOCKET MATCHED\n");
#endif
		return;
	} else {
#if DEBUGUDP
		printf("SOCKET MATCHED %i ", sock);
#endif
		/* 3a: If there is no room, black hole the packet. */
		if((MAXTAILR - rbufcount[sock]) < totallen) {
#if DEBUGUDP
			printf("NO ROOM ");
#endif
			goto black_hole;
		}
#if DEBUGUDP
		printf("SAVING PACKET ");
#endif
		/* 3b: Place package into the queue. */
		bufferdatamove(&buffer[9], sock, 1); /* IP information */
		bufferdatamove(&buffer[12], sock, 4); /* IP information */
		bufferdatamove(&buffer[iplen], sock, udplen); /* UDP information */
	}

black_hole:
	freebuffer(bufnum);
#if DEBUGUDP
	printf("DONE.\n");
#endif
	return;
}


#else /* UDP_ENABLED */
void udp(bufnm, iplen, totallen, udplen)
unsigned int bufnum;
unsigned int iplen;
unsigned int totallen;
unsigned int udplen;
{
	/* type 3, code 3 protocol unreachable */
	err_reply(bufnum, 3, 2, 0);
	return;
}
#endif /* UDP_ENABLED */

#else /* IP_ENABLED */
void udpnotused(VOIDFIX)
{
	return(0);
}
#endif /* IP_ENABLED */
/* tell link parser who we are and where we belong

0600 libznet.lib udp.obj xxLIBRARYxx

*/

